/**
 * Chat room controller.
 *
 *
 * Message states:
 *
 *    on the server side message can be in one the 5 states and they are displayed a bit differently (*) on the UI:
 *
 *    SERVER:           UI:
 *    - sent            - sending *
 *    - queued          - sent    *
 *    - delivered       - delivered
 *    - read            - read
 *    - deleted
 *
 *
 * Media asset states (NOT IMPLEMENTED YET):
 *
 *    - uploading
 *    - uploaded    , when the asset has been uploaded to the server
 *    - viewed      , means that recipient viewed the full-size version of the image
 *
 */

(function (define) {
  'use strict';

  define(['config', 'utils'], function (config, utils) {

    var _ = utils._,
      maybe = utils.maybe({});

    var ChatRoomCtrl = function (
      $scope, $rootScope, $stateParams, $q, $timeout, $ionicPopover, baseData, $location, $sceDelegate, $sce, $translate,
      uuid4, $ionicScrollDelegate, transportBroker, appState, dialogs,
      messageCollection, chatsCollection, chatsArchivedData, userProfile, chatPlusDialog, fullScreenDialog, notificationPopover, security
    ) {
      ionic.keyboard.hide();

      var _cacheCleared = false,
        myProfile = userProfile.getFull(),
        myProfileId = $scope.myProfileId = myProfile.profileId,
        currentChatId = $stateParams.chatId,
        currentChat,
        isChatActive,
        currentChatCollection,
        participant,
        lastSentTypingTime = 0,
        blockUpdate = false,
        syncOff;

      var collection = messageCollection(currentChatId, myProfileId);

      // link messages and collection items:
      $scope.messages = collection.items();
      console.log('ChatRoom: started with messages: ' + $scope.messages.length + ': ' + _.map(_.prop('text'), $scope.messages).join(', '));
      scrollToBottom(false, 10);
      messagesAsRead($scope.messages);

      if ($scope.messages.length === 0 && !collection.wasLoaded()){
        $scope.isLoading = true;
      }

      syncCollection();

      collection.onChange($scope, function(data){
        console.log('[ChatRoomCtrl.onChange] ' + data.type);
        messagesAsRead($scope.messages);
        switch(data.type){
          case 'new-item-mine':
            scrollToBottom(false);
            break;
          case 'new-item':
            scrollToBottom(true, 1);
            break
          case 'item-state':
            $scope.$apply();
            break;
        }
      });

      onChatReady();
      onCollectionUpdate();
      //transportBroker.joinChat(currentChatId);

      function syncCollection() {
        //console.log('[.syncCollection] ...');
        syncOff && (console.log('syncOff()'), syncOff(), syncOff = null);
        $scope.messages.length < 10 && collection.sync(null, $scope.messages.length).then(onSync).catch(onSyncErr);
      }
      function onSync() {
        //console.log('ChatRoom: sync done');
        $scope.isLoading = false;
        onCollectionUpdate();
        scrollToBottom(false, 100);
        messagesAsRead($scope.messages);
      }
      function onSyncErr(e) {
        //console.log('[onSyncErr] ' + JSON.stringify(e));
        syncOff = $scope.$on('app-online', syncCollection);
      }
      function onChatReady(){
        currentChat = chatsCollection.get(currentChatId) || chatsArchivedData.get(currentChatId);
        isChatActive = !!chatsCollection.get(currentChatId);
        currentChatCollection = isChatActive ? chatsCollection : chatsArchivedData;
        participant = currentChat && currentChat.participant;
        participant.isNewMatch = currentChat.isNewMatch;
        participant.lastMessageTimeStamp = currentChat.lastMessageTimeStamp;

        $scope.participant = participant;
        $scope.otherParticipantProfileId = participant && participant.profileId;
        $scope.title = $scope.participantName = participant && participant.firstName;
        $scope.isChatActive = isChatActive;
        $scope.videoPlaying = false;
        $scope.audioPlaying = false;
      }
      function onCollectionUpdate() {
        $scope.hasMoreMessages = collection.hasMoreMessages();
        $scope.isNewMatch = collection.isNewMatch();
        $scope.createdTimestamp = collection.createdTimestamp();
      }

      if (!currentChat) {
        chatsCollection.load().then(onChatReady);
      }

      $scope.myProfileId = myProfileId;
      $scope.myProfile = myProfile;
      $scope.currentChatId = currentChatId;
      //$scope.otherParticipantProfileId = 'kismet_support_profile'; // default to hide button off the start
      $scope.error;
      $scope.timeoutPromise;
      $scope.showDelete = false;
      $scope.keyboardFocused = true;
      $scope.closeKeyboard = function(){
        $scope.keyboardFocused = false;
        if (window.cordova && window.cordova.plugins.Keyboard) cordova.plugins.Keyboard.close();
      };

      $scope.stayOpen = false;

      $scope.$on('app-resume', function(){
        // TODO: disabling for now since app pause/resume is emitted when user picks a picture to send, and we dont want to refresh message collection in this case.
        //clearCache();
        //collection.sync();
        //messagesAsRead($scope.messages);
      });

      $scope.$on('screenshot-taken', function(){
        baseData.post('/api/chat/screen-capture?profileId={profileId}&chatId=' + currentChatId, null, userProfile.getProfileId());
      });

      window.addEventListener('native.keyboardshow', keyboardHandler);
      window.addEventListener('native.keyboardhide', keyboardHandler);
      notificationPopover.listenAndPopup($scope, null, {beforeCheck: function(data){ return data.chatId !== $scope.currentChatId;}});

      /***  Init steps  ***/

      $ionicPopover.fromTemplateUrl('templates/dialogs/chat-room-popover.html', {
        scope: $scope
      }).then(function(popover) {
        $scope.popover = popover;
        $scope.wipeThisChat = function(){
          dialogs.wipeChat().then(function(){
            $scope.popover.hide();
            wipeThisChat();
            collection.clear();
          });
          /*$scope.showDeleteOptions = function(){
            $scope.showDelete = true;
            $scope.popover.hide();
          }*/
        };
        $scope.blockUser = function(){
          dialogs.blockUser().then(function(){
            return chatsCollection.blockUser(currentChatId, myProfileId);
          }).then(function(){
            collection.clear();
            collection.wipe();
            $location.path('/app/chats/active');
          }).catch(function() {
            $scope.popover.hide();
          });
        };
        $scope.viewProfile = function() {
          $scope.goto('/app/profiles/' + $scope.otherParticipantProfileId);
        }
      });

      /***  View exposed methods  ***/
      $scope.sendMessage = function() {
        $timeout(function(){
          $scope.stayOpen = true;
          //console.log("* KEYBOARD * : CTRL stayOpen set true");
        }, 50);

        var messageId = uuid4.generate(),
          text = $scope.inputMessage;

        if (!text){
          console.log('WARNING: empty message !!! ' + messageId);
          return;
        }

        var messageItem = collection.tryAdd({
          messageId: messageId,
          text: text,
          from: $scope.myProfileId,
          isMine: true,
          state: 'sent'
        });
        updateChat(messageItem);

        $scope.inputMessage = '';
      };
      $scope.removeMessage = function(message) {
        if (message.participant.profileId == $scope.myProfileId) {
          collection.tryRemove(message);
        }
      };
      var wipeDialogOpen = false;
      $scope.askRemove = function(message){
        if (message.participant.profileId == $scope.myProfileId && !wipeDialogOpen) {
          wipeDialogOpen = true;
          navigator.vibrate(config.popups.confirmVibrateTime);
          if (window.cordova && window.cordova.plugins.Keyboard) cordova.plugins.Keyboard.close();
          dialogs.wipeMessage()
              .then(function(){
                wipeDialogOpen = false;
                $scope.removeMessage(message);
              })
              .catch(function(){
                wipeDialogOpen = false;
              });
        }
      };
      $scope.typing = function(){
        if (lastSentTypingTime == 0) {
          transportBroker.typing(currentChatId, $scope.myProfileId);
          lastSentTypingTime = new Date();
        } else {
          var endTime = new Date();
          if ((endTime - lastSentTypingTime)/1000 > 2) {
            lastSentTypingTime = 0;
          }
        }
      };
      // top-right *** menu button
      $scope.popoverOptions = function($event){
        $scope.popover.show($event);
      };
      $scope.$on('$destroy', function() {
        $scope.popover.remove();
      });

      /***  Message manipulation functions  ***/
      function removeMessage(message, whenDone) {
        message.isRemoved = true;
        clearCache();
        $timeout(function(){
          collection.remove(message.messageId);
          whenDone && whenDone();
        }, 200);
      }
      function messagesAsRead(messages) {
        if (appState.get('isBackground')) return;
        var messagesToRead = _.compose(_.map(messageAsRead), _.filter(function(m){ return !m.isMine && m.state !== 'read';}))(messages);
        if (messagesToRead.length > 0) {
          clearCache();
          chatsCollection.updateChat({chatId: currentChatId, hasNewMessage: false});
        }
      }
      function messageAsRead(message) {
        if (appState.get('isBackground')) return;
        if (message.state === 'downloading') {
          collection.messageAsRead(message);
        }
        if (message.state === 'queued' || message.state === 'delivered' || message.state === 'sent') {
          collection.messageAsRead(message);
        }
      }
      function wipeThisChat() {
        transportBroker.wipeChat(currentChatId, $scope.myProfileId);
        wipeMyMessages();
      }
      function wipeMyMessages() {
        var count = 0, deferreds = [];
        $scope.messages.filter(function(item){ return item.isMine;}).reverse().forEach(function(item){
          var deferred = $q.defer();
          $timeout(function(){
            removeMessage(item, function(){
              deferred.resolve('OK');
            });
          }, (count++)*300);
          deferreds.push(deferred.promise);
        });
        $q.all(deferreds).then(function(){
          collection.clear();
          collection.sync();
          $scope.doRefresh().then(function(){
            scrollToBottom(true, 1000);
          });
        });
      }
      function clearCache() {
        if (!_cacheCleared) {
          //console.log('ChatRoom: Will need to clear cache for data collections.');
          chatsCollection.clear();
          chatsArchivedData.clear();
          collection.clear();
          _cacheCleared = true;
        }
      }
      function updateChat(message) {
        if (!$scope.isChatActive) {
          unarchiveChat(currentChatId);
          $scope.isChatActive = true;
        }
        // chatId, lastMessage, lastMessageFromYou, lastMessageTimeStamp, hasNewMessage
        chatsCollection.updateChat({
          chatId: currentChatId,
          lastMessage: message.text,
          lastMessageFromYou: message.isMine,
          lastMessageTimeStamp: message.timestamp,
          hasNewMessage: false
        });
      }
      function unarchiveChat(chatId) {
        var chat = chatsArchivedData.remove(chatId);
        chatsCollection.add(chat);
      }


      /***  Other handlers  ***/
      function keyboardHandler(){
        //console.log($ionicScrollDelegate.$getByHandle('chatRoomScroll').getScrollPosition().top + " " + $ionicScrollDelegate.$getByHandle('chatRoomScroll').getScrollView().getScrollMax().top);
        if ($ionicScrollDelegate.$getByHandle('chatRoomScroll').getScrollPosition().top+50 >= $ionicScrollDelegate.$getByHandle('chatRoomScroll').getScrollView().getScrollMax().top) {
          scrollToBottom(true, 300);
        }
      }
      function scrollToBottom(animate, timeoutValue) {
        animate = animate || false;
        timeoutValue = timeoutValue || 0;

        if (timeoutValue) {
          $timeout(function () {
            $ionicScrollDelegate.$getByHandle('chatRoomScroll').scrollBottom(animate);
          }, timeoutValue);
        } else {
          $ionicScrollDelegate.$getByHandle('chatRoomScroll').scrollBottom(animate);
        }
      }
      function getTrustedUrl(mediaId) {
        var mediaUrl = config.getMediaUrl(mediaId);
        return $sce.trustAsResourceUrl(mediaUrl + security.bewit(mediaUrl, 6*600*1000));
      }

      function getUntrustedUrl(mediaId) {
        var mediaUrl = config.getMediaUrl(mediaId);
        return mediaUrl + security.bewit(mediaUrl, 6*600*1000);
      }

      // Bottom-left PLUS button:
      $scope.chatRoomOptions = function(){
        var mediaId = uuid4.generate();

        blockUpdate = true;
        chatPlusDialog
          .popup({chatId: currentChatId, mediaId: mediaId})
          .then(function(data){
            console.log('ChatRoomCtrl.chatRoomOptions: popup resolved with: ', data);
            var messageId = uuid4.generate();
            var messageData = {
              messageId: messageId,
              from: $scope.myProfileId,
              isMine: true
            };
            if (data.mediaType === 'location') {
              // TODO: send to mqtt as regularily.
              messageData.location = data.location;
              collection.tryAdd(messageData);
              return;

            } else {
              messageData.state = 'uploading';
              messageData.digitalAssets = [{
                //id: mediaId,
                type: data.mediaType,
                state: 'uploading',
                base64: data.base64
              }];
            }
            // add to collection, but don't send to mqtt:
            console.log('- add local message, uploading...');
            var messageItem = collection.tryAdd(messageData, 'no-request');

            // This is the deferred upload action:
            if (data.deferredUpload) {
              data.deferredUpload.then(function(digitalAsset){
                onMediaUploaded(messageItem, digitalAsset);
              }).catch(function(err){
                onUploadFailure(messageItem, data.retryAction, mediaId, err);
              }).finally(function(){
                blockUpdate = false;
              });
            } else {
              blockUpdate = false;
            }
          })
          .catch(function(error){
            // Cancelled by user. No action here.

            if (error && error.mediaType) {
              switch (error.mediaType) {
                case 'location':
                  $rootScope.$broadcast('app-location-error');
                  $timeout(function(){
                    $rootScope.$broadcast('app-location-good');
                  }, 3000);
                  break;
              }
            }

            blockUpdate = false;
            //alert('chatRoomOptions: ERROR ' + error + '. ' + JSON.stringify(error,1).substring(0,350));
          });
      };
      function onMediaUploaded(messageItem, digitalAsset) {
        //messageItem.digitalAssets[0].src = digitalAsset.id;
        messageItem.digitalAssets[0].src = getTrustedUrl(digitalAsset.id);
        messageItem.digitalAssets[0].id = digitalAsset.id;
        messageItem.digitalAssets[0].state = 'uploaded';
        messageItem.digitalAssets[0].duration = digitalAsset.duration;
        messageItem.state = 'sent';
        messageItem.text = messageItem.textHtml = '';
        console.log('- media uploaded, sending message to server... digitalAsset, messageItem:', digitalAsset, messageItem);
        collection.tryAddRequest(messageItem);
      }
      function onUploadFailure(messageItem, retryAction, mediaId, err) {
        console.log('onUploadFailure: ' + JSON.stringify(err));
        // TODO: retry action here:
        messageItem.state = 'failed';
        messageItem.displayState = 'upload failed';
        var asset = messageItem.digitalAssets && messageItem.digitalAssets[0];
        asset && (asset.state = 'failed');
        !messageItem.retry && (messageItem.retry = function(){
          console.log('- retrying media upload for: ' + mediaId);
          messageItem.state = 'uploading';
          messageItem.displayState = 'uploading';
          asset && (asset.state = 'uploading');
          retryAction().then(function(digitalAsset){
            onMediaUploaded(messageItem, digitalAsset);
          }).catch(function(err){
            onUploadFailure(messageItem, retryAction, mediaId, err);
          });
        });
      }

      function stopMediaPlayback()
      {
        //if ($scope.videoPlaying) {
        //  $scope.videoPlaying.pause();
        //  $scope.videoPlaying = null;
        //}

        if ($scope.audioPlaying && document.getElementById($scope.playingAudioId))
        {
          $scope.audioPlaying.pause();
          $scope.audioPlaying.currentTime = 0;
          $scope.audioPlaying = null;
        }
      }

      $scope.playVideo = function(asset)
      {
        VideoPlayer.play(asset.srcNatural);
        //stopMediaPlayback();
        //
        //if(asset.src == $scope.playingVideoId)
        //{
        //  $scope.playingVideoId = '';
        //}
        //else
        //{
        //  //Clicked on another video while the first is still playing..therefore stop first and play second
        //  var video = document.getElementById(asset.src);
        //  if (video != null)
        //  {
        //    video.onended = function(e) {
        //      $scope.playingVideoId = '';
        //    };
        //
        //    $scope.videoPlaying = video;
        //    $scope.playingVideoId = asset.src;
        //    $scope.videoPlaying.play();
        //  }
        //}
      }

      $scope.playAudio = function(asset)
      {
        if ($scope.playingAudio != null && $scope.playingAudio.src === asset.srcNatural)
        {
          $scope.playingAudio.pause();
          $scope.playingAudio = null;
        }
        else if ($scope.playingAudio != null && $scope.playingAudio.src != asset.srcNatural)
        {
          $scope.playingAudio.pause();

          var myMedia = new Media(asset.srcNatural);
          $scope.playingAudio = myMedia;
          myMedia.play();
        }
        else
        {
          var myMedia = new Media(asset.srcNatural, function()
          {
            //The callback that executes after a Media object has completed the current play, record, or stop action. (Function)
            $scope.playingAudio.stop();
            $scope.playingAudio.release();
            $scope.playingAudio = null;
          });
          $scope.playingAudio = myMedia;
          myMedia.play();
        }


      }

      $scope.showFullImage = function(options){

        // TODO Crete only one modal and reuse it not only for all image messages but for all chats
        var modalScope = $rootScope.$new(true, $scope);
        modalScope.imageId = options && options.imageId;
        modalScope.src = options && options.src;
        fullScreenDialog
          .popup({
            type: 'image',
            modalScope: modalScope
          })
          .then(function(modal){
            modalScope.modal = modal;
            modalScope.hide = function(){
              modalScope.modal.hide();
            };
            modal.show();
          });
        modalScope.$on('$destroy', function() {
        });
        // Execute action on hide modal
        modalScope.$on('modal.hidden', function() {
        });
        // Execute action on remove modal
        modalScope.$on('modal.removed', function() {
          // Execute action
        });
      };
      $scope.doRefresh = function() {
        if (!$scope.hasMoreMessages){
          $scope.$broadcast('scroll.refreshComplete');
          return;
        }
        var lastMessageId = $scope.messages[0] && $scope.messages[0].messageId;
        return collection
          .sync(lastMessageId)
          .then(function(messages){
            //console.log('loadMessages for ' + lastMessageId, messages);
            $scope.$broadcast('scroll.refreshComplete');
            $scope.hasMoreMessages = collection.hasMoreMessages();
          });
      };
      $scope.audioSend = function(data) {
        $scope.isAudioShown = false;
        //console.log('audioSend: ', data);
        var messageId = uuid4.generate(),
          mediaId = uuid4.generate();

        var messageData = {
          messageId: messageId,
          text: '',
          from: $scope.myProfileId,
          isMine: true,
          state: 'uploading',
          digitalAssets: [{
            //id: mediaId,
            type: 'audio',
            state: 'uploading'
          }]
        };
        var messageItem = collection.tryAdd(messageData, 'no-request');

        function uploadAudio() {
          return data.upload(data.fileEntry, {mediaId: mediaId, chatId: currentChatId, mediaType: 'audio'});
        }

        uploadAudio().then(function(digitalAsset){
          console.log('-');
          onMediaUploaded(messageItem, digitalAsset);
        }).catch(function(err){
          onUploadFailure(messageItem, uploadAudio, mediaId, err);
        });
      };
      $scope.audioCancel = function(){
        $scope.isAudioShown = false;
        //console.log('audioCancel.');
      };
    };

    return ChatRoomCtrl;
  });

}(define));
