(function (define) {
  'use strict';

  define(['config', 'data/data-mock'], function (config, dataMock) {

    var baseData = function ($q, $http, $rootScope, security) {

      var _get, _post, _getCached,
        apiPrefix = function() { return config.server.url; };

      _get = function(url, profileId, noPing) {
        var deferred = $q.defer(),
          fullUrl = (apiPrefix() + url).replace('{profileId}', (profileId || ''));

        $http({
          method: 'GET',
          url: fullUrl,
          headers: security.header(fullUrl, 'GET')
          //withCredentials: true
        })
          .success(function(response){
            if (response.header && response.header.errorCode) {
              console.log('GET ERROR: ' + response.header.errorCode + ' ' + response.header.errorMessage);
              deferred.reject(response.header);
            } else {
              deferred.resolve(response.payload);
            }
          })
          .error(function(data, status, headers, config){
            deferred.reject(data);
            if (data.statusCode == 401 && !noPing) {
              ping(fullUrl);
            }
          });

        return deferred.promise;
      };

      _post = function(url, data, profileId, noPing) {
        var deferred = $q.defer(),
          fullUrl = (apiPrefix() + url).replace('{profileId}', (profileId || ''));
        $http({
          method: 'POST',
          url: fullUrl,
          data: data,
          headers: security.header(fullUrl, 'POST')
          //withCredentials: true
        })
          .success(function(response){
            deferred.resolve(response.payload);
          })
          .error(function(data, status, headers, config){
            deferred.reject(data);
            if (data && data.statusCode == 401 && !noPing) {
              ping(fullUrl);
            }
          });

        return deferred.promise;
      };

      function ping(fullUrl){
        var url = config.apiUrls.PING;
        return _get(url, null, true).then(function(){
          //console.log('PING OK after failure to ' + fullUrl);
        }).catch(function(data){
          //console.log('PING FAILED after failure to ' + fullUrl);
          if (data.statusCode == 401) {
            $rootScope.$broadcast('forced-logout');
          }
        });
      }


      /**
       * Cache store
       * @key: <url | string>
       * @item:
       *  {
       *    url: <string>,
       *    ttl: <number of sec | -1 for unlimit>,
       *    timestamp: <timestamp>,
       *    promise: <$q.defer.promise>
       *  }
       */
      var cacheStore = {};

      $rootScope.$on('logout', function(){
        cacheStore = {};
      });

      /**
       * Caches GET request for <ttl> seconds
       * @param url
       * @param opt	{
       * 		ttl: <number of sec, -1 means forever, 0 means don't cache>,
       * 		cacheId,
       * 	  profileId
       * }
       * @returns {<ajax promise>}
       */
      _getCached = function(url, opt) {
        //for legacy:
        var profileId = opt && typeof opt === 'string' && opt || opt && opt.profileId;

        var cacheId = opt && opt.cacheId || url,
          ttl = opt && typeof opt.ttl !== 'undefined' ? opt.ttl : config.requestTtl,
          cacheItem = cacheStore[cacheId],
          lastUpdated = cacheItem && cacheItem.timestamp,
          lastTtl = cacheItem && cacheItem.ttl,
          now = (new Date).getTime(),
          diff = parseInt((now - lastUpdated) / 1000),
          revalidate = !lastUpdated || (ttl === 0 || lastTtl === 0) || (ttl !== -1 && ttl < diff);

        if (revalidate) {
          //console.log('_getCached: revalidate for ' + url.substring(0,25) + ', ttl=' + ttl + ', diff=' + diff + ', now=' + now + ', lastUpdated=' + lastUpdated + ', opt.ttl=' + (opt && opt.ttl), url);
          cacheStore[cacheId] = {
            url: url,
            ttl: ttl,
            timestamp: (now || (new Date).getTime()),
            promise: _get(url, profileId)
          };

          if (opt && opt.newDataEvent){
            //console.log('BaseData: newDataEvent=' + opt.newDataEvent);
            cacheStore[cacheId].promise.then(function(payload){
              //console.log('BaseData: broadcasting newDataEvent=' + opt.newDataEvent);
              $rootScope.$broadcast(opt.newDataEvent, payload);
            });
          }
        } else {
          //console.log('_getCached: from cache ' + url.substring(0,25) + ', ttl left: ' + (ttl - diff), url);
        }

        return cacheStore[cacheId].promise;
      };

      var clearCache = function(cacheId) {
        //console.log('BaseData.clearCache for ' + cacheId);
        if (cacheStore[cacheId]){
          delete cacheStore[cacheId];
          //console.log('BaseData.clearCache done');
        } else {
          //console.log('BaseData.clearCache: no cache for the cacheId', cacheStore);
        }
      };

      var resetCacheTtl = function(cacheId) {
        //console.log('BaseData.resetCacheTtl for ' + cacheId);
        if (cacheStore[cacheId]){
          cacheStore[cacheId].ttl = 0;
          //console.log('BaseData.resetCacheTtl done');
        } else {
          //console.log('BaseData.resetCacheTtl: no cache for the cacheId');
        }
      };

      return {
        _get: _get,
        get: _getCached,
        post: _post,
        clearCache: clearCache,
        resetCacheTtl: resetCacheTtl,
        ping: ping
      }
    };

    return baseData;
  })
}(define));