/**
 * Push Notifications
 * @param options {
 *    senderId: <>,
 *    deviceID: <deviceId>,
 *    userPlatform: <string>,
 *    eventTriggerDelegate: <function>,
 *    registerDelegate: <function>,
 *    localStoragePrefix: <string>
 * }
 */


(function (define) {
  'use strict';

  define(['config'], function (config) {

    var errorLog = {error: function(a){ console.log('ERROR: ' + a);} },
      debug = {alertMsg: function(a, o){ console.log(a + ': ' + JSON.stringify(o,1));}, toConsole: function(a,b){console.log(a,b);}};

    var pushData = function ($rootScope, $location, $timeout, baseData, userProfile) {
      var _regId,
        profileId,
        registerUrl = '/api/profile/register-push?profileId={profileId}',
        EVENTS = {
          newMessage: 'push-message'
        },
        defaultEventTriggerDelegate = function(eventName, data){
          $rootScope.$broadcast(eventName, data);
        },
        defaultRegisterDelegate = function(data) {
          return userProfile.getLoginInfo()
            .then(function(loginInfo){
              var url = registerUrl.replace('{profileId}', loginInfo.profileId);
              debug.alertMsg('Trying to register on server via ' + url);
              profileId = loginInfo.profileId;
              return baseData.post(url, data);
            })
            .then(function(data){
              debug.alertMsg('pushData: registered on server' + JSON.stringify(data,1).substring(0,300));
            })
            .catch(function(error){
              var url = registerUrl.replace('{profileId}', profileId);
              debug.alertMsg('ERROR: cannot register device for push via ' + url, error);
            });
        };

      var push = {
        pushNotification : null,
        regId: null,
        initialize: function(options) {
          push.options = angular.extend({
            senderId: config.push.senderId,
            deviceID: null,
            userPlatform: null,
            localStoragePrefix : 'KismetPush',
            eventTriggerDelegate: defaultEventTriggerDelegate,
            registerDelegate: defaultRegisterDelegate
          }, options);

          debug.alertMsg('push.initialize', options);
          debug.alertMsg('push.initialize ' + device.platform);

          try {
            push.pushNotification = window.plugins.pushNotification;
            if (device.platform == 'android' || device.platform == 'Android' || device.platform == 'amazon-fireos' ) {
              if (!window.onNotification) {
                window.onNotification = push.onNotification;
              }
              debug.alertMsg('about to register for GCM...');
              push.pushNotification.register(push.successHandler, push.errorHandler, {"senderID": push.options.senderId,"ecb":"onNotification"});		// required!
            } else {
              if (!window.onNotificationAPN) {
                window.onNotificationAPN = push.onNotificationAPN;
              }
              push.pushNotification.register(push.tokenHandler, push.errorHandler, {"badge":"true","sound":"true","alert":"true","ecb":"onNotificationAPN"});	// required!
            }
          } catch(err) {
            debug.alertMsg('PUSH ERROR');
            var txt="There was an error on this page.\n\n";
            txt+="Error description: " + err.message + "\n\n";
            debug.alertMsg(txt);
            errorLog.error({message: 'Cannot initialize push notifications', error: err});
          }
        },
        // handle APNS notifications for iOS
        onNotificationAPN: function(e) {
          /*
           Data e: {
             text: <actual message text>,
             badge: <number>,
             alert: <the same message text>,
             chatId: <id>,
             otherProfileId: <id>,
             event: <"newChatMessage">,
             otherFirstName: <name>,
             otherImageId: <id>,
             foreground: <"0">
           }
           */

          if (e.alert) {
            push.triggerMessageEvents(e.event, e);
          }

          if (e.sound) {
            // playing a sound also requires the org.apache.cordova.media plugin
            var snd = new Media(e.sound);
            snd.play();
          }

          if (e.badge) {
            //push.pushNotification.setApplicationIconBadgeNumber(push.successHandler, e.badge);
          }
        },

        // handle GCM notifications for Android
        onNotification: function(e) {
          debug.alertMsg('pushData: onNotification for GCM: ' + JSON.stringify(e));
          switch ( e.event ) {
            case 'registered':
              if ( e.regid.length > 0 ) {
                debug.alertMsg('pushData: registered in GCM: ' + e.regid);
                push.saveRegId(e.regid);
              }
              break;

            case 'message':
              /*
               Data e: {
                 event: 'message',
                 collapse_key,
                 from,
                 foreground: <boolean>,
                 message: <text>,
                 msgcnt: <string number>,
                 payload: { json: {
                   text, otherImageId, otherFirstName, chatId, otherProfileId,
                   event: <"newMatch" | "newChatMessage" | "systemMessage">
                 }},
               }
               */
              var data = e && e.payload && e.payload.json,
                eventType = data && data.event;		// { "newMatch" | "newChatMessage" | "systemMessage" }

              data.foreground = e.foreground;

              push.triggerMessageEvents(eventType, data);
              break;

            case 'error':
              errorLog.error({message: 'Push event error: ' + e.msg, e: e});
              debug.alertMsg("ERROR:" + e.msg);
              break;

            default:
              errorLog.error({message: 'Push: unknown event received: ' + e.event, e: e});
              debug.alertMsg("Unknown, an event was received and we do not know what it is");
              break;
          }
        },

        tokenHandler: function(result) {
          push.saveRegId(result);
        },

        successHandler: function(result) {
          //debug.toConsole("Push register success: " + result);
        },

        errorHandler: function (error) {
          errorLog.error({message: 'Push register error', error: error});
          debug.alertMsg("Push register error", error);
        },

        saveRegId: function (regId) {
          _regId = regId;
          debug.alertMsg('saveRegId: ' + regId.length);
          push.registerPush(regId, function (data) {
            //errorLog.info({message: 'Successfully registered regId', regId: regId, response: data});
            localStorage.setItem(push.options.localStoragePrefix + 'pushRegId', regId);
          });
        },
        triggerMessageEvents: function(eventType, data) {
          if (!profileId) profileId = userProfile.getProfileId();
          // eventType: { "newMatch" | "newChatMessage" | "systemMessage" }
          console.log('push.onNotification: eventType=' + eventType + ', data: ' + JSON.stringify(data,1));
          if (data.toProfileId && data.toProfileId !== profileId) {
            console.log('WARNING: received push for a foreign user (' + data.toProfileId + ')');
            return;
          }
          push.options.eventTriggerDelegate(EVENTS.newMessage, data);
          var obj;
          if (eventType === 'newChatMessage') {
            // data: {otherImageId, otherFirstName, text, messageId, chatId, otherProfileId}
            if (data.otherProfileId === profileId) {
              console.log('WARNING: received push for the user himself', data);
              return;
            }
            if (!data.foreground || data.foreground === "0") {
              console.log('PUSH: received in background. Navigating to chat room... data.chatId=' + data.chatId);
              $timeout( function(){
                $location.path('/app/chat-room/' + data.chatId);
              }, 100);
              return;
            } else {
              console.log('PUSH: received in foreground. Do NOT navigate to chatroom.');
            }
            obj = {
              chatId: data.chatId,
              messageId: data.messageId,
              text: data.text,
              from: {
                profileId: data.otherProfileId,
                firstName: data.otherFirstName,
                imageId: data.otherImageId
              },
              digitalAssets: data.digitalAssets
            };
            push.options.eventTriggerDelegate(config.event.CHAT_MESSAGE_NEW, obj);
          }
          if (eventType === 'newMatch') {
            //alert('PUSH: eventType=' + eventType + ', data: ' + JSON.stringify(data,1));
            push.options.eventTriggerDelegate(config.event.NEW_MATCH, {
              text: 'New match! ' + JSON.stringify(data, 1),
              participants: [
                {imageId: data.imageId, firstName: 'You', isYou: true},
                {imageId: data.otherImageId, firstName: data.otherFirstName, profileId: data.otherProfileId, chatId: data.chatId}
              ]
            });
          }
        },
        registerPush: function (regId, onSuccess) {
          var data = {
            deviceId: push.options.deviceID,
            platform: push.options.userPlatform,
            pushToken: regId
          };
          debug.alertMsg('registerPush: ' + JSON.stringify(data, 1));
          //server.sendPostWithPid('/api/profile/register-push', data, onSuccess, push.onRegisterPushErr);
          push.options.registerDelegate(data);
        },
        onRegisterPushErr: function (data) {
          debug.alertMsg('onRegisterPushErr', data);
        }
      };

      return {
        init: push.initialize
      };
    };

    return pushData;
  });
}(define));