
(function (define, io) {
  'use strict';

  define(['config', 'utils'], function (config, utils) {

    var _ = utils._;

    var websocketData = function ($rootScope, $q, userProfile, security, baseData) {

      var websocket, userData, ready,
        wsUrl = function() { return config.server.url + ':' + config.server.webSocketPort;},
        messageQueue = [];

      var init = function() {
        //console.log('Websocket init...');
        if (!ready) {
          ready = $q.defer();
        }
        websocket = io.connect(wsUrl(), {forceNew: true, query: 'signature=' + security.getSignature()});

        //console.log('Websocket created for ' + wsUrl());

        websocket.on('connect', function () {
          console.info('Successfully established a working connection \\o/');
          runSendingJob();
          fire('ws-connect');
          ready.resolve('Websocket ready');
        });
        websocket.on('message', function(msg) {
          onMessage(msg);
        });
        websocket.on('disconnect', function(){
          //console.log('ws disconnect :(');
          fire('ws-disconnect');
        });
        websocket.on('reconnect', function(){
          //console.log('ws reconnect ...');
          runSendingJob().finally(function(){
            fire('ws-reconnect');
          });
        });
        websocket.on('error', function (reason){
          console.error('Unable to connect Socket.IO', reason);
          baseData.ping('Websocket');
        });

        userData = userProfile.getFull();
        //console.log('Websocket init done. User data:', userData);
      };
      var destruct = function() {
        //console.log('Websocket destruct...');
        disconnect();
        if (ready){
          ready.reject('websocket destructing on logout...');
          ready = null;
        }
        if (userData)  userData.profileId = null;
      };

      var disconnect = function() {
        websocket && websocket.close();
      };

      //$rootScope.$on('app-pause', disconnect);
      //$rootScope.$on('app-resume', init);

      function onMessage(msg) {
        console.log('<<<<<<   ws-on-message: ' + (msg.scope || '-') + ' : ' + msg.type + ' : ' + (msg.payload && msg.payload.text || '-'), msg);
        /*
         Data structure: {
           scope: <'message' | 'chat' | 'app'>,
           type: <'sent' | 'joint' | 'left' | 'typing'>,
           chatId,
           messageId,
           from: <{profileId, firstName} | profileId>,
           payload: {text: <string>}
         }
         */
        if (msg.type === 'queued') {

          // acknowledgement:
          if (msg.from === userData.profileId) {
            fire('ws-queued-' + msg.chatId, msg);
          }
          // new message:
          else {
            var obj = {
              chatId: msg.chatId,
              messageId: msg.messageId,
              text: msg.payload && msg.payload.text,
              from: msg.from && msg.from.profileId && msg.from || {profileId: msg.from},
              location: msg.payload.location
            };
            if (msg.payload.digitalAsset) {
              obj.digitalAssets = [msg.payload.digitalAsset];
            }
            fire(config.event.NEW_CHAT_MESSAGE + '-' + msg.chatId, obj);
          }
        }

        /*
         { chatId:"2222222222", from:"1111111111", fromFirstName:"TheFirstName", fromImageId:"111-222-333-444", scope:"chat", type:"new" }
         */
        if (msg.scope === 'chat' && msg.type === 'new') {
          fire(config.event.NEW_MATCH, {
            participants: [
              {imageId: msg.fromImageId, firstName: msg.fromFirstName, profileId: msg.from, chatId: msg.chatId}
            ]
          });
        }
        if (msg.type === 'deleted' || msg.type === 'delete') {
          fire('ws-message-deleted', msg);
        }
        if (msg.type === 'typing') {
          fire('ws-typing', msg);
        }
        if (msg.type === 'wipe') {
          fire('ws-wipe', msg);
        }
        if (msg.type === 'read') {
          //{chatId: "543d33c2f9ff4f282b1d6d5d", from: {firstName, profileId}, scope: "message", messageId: "2c351be8-47db-4fed-bbca-ebd9a0d1dea4", type: "read"}
          fire('ws-read', msg);
        }
        if (msg.type === 'uploaded') {
          //{chatId, payload: {digitalAsset: {id, type, contentType}}}
          fire('ws-media-uploaded', msg);
        }
      };

      var send = function(messageId, chatId, scope, type, payload, from) {
        var obj = {
          chatId: chatId,
          scope: scope,
          from: from && userData.profileId || {
            profileId: userData.profileId,
            firstName: userData.firstName
          },
          type: type,
          timestamp: moment(),
          messageId: messageId,
          payload: payload
        };

        //console.log('ws added to the queue: ', obj);
        messageQueue.push(obj);
        runSendingJob();
      };

      var isJobRunning = false;
      function runSendingJob() {
        if (!isJobRunning && websocket && websocket.connected && messageQueue.length > 0){
          isJobRunning = true;
          return sendPendingMessages().finally(function(){
            isJobRunning = false;
            runSendingJob();
          });
        }
        return $q.reject('empty');
      }

      function sendPendingMessages() {
        var deferreds = [];
        if (websocket && websocket.connected && messageQueue.length > 0) {
          //console.log('ws messageQueue: sending ' + messageQueue.length);
          messageQueue.forEach(function(obj){
            var deferred = $q.defer();
            deferreds.push(deferred.promise);
            console.log('> >> >   ws-send: ' + obj.scope + ' : ' + obj.type + ' : ' + (obj.payload && obj.payload.text || '-'), obj);
            obj._try = ++obj._try || 1;
            websocket.json.send(obj, function(){
              //console.log('ws ACK : sent after ' + obj._try + ', ' + obj.scope + ' : ' + obj.type + ' : ' + (obj.payload && obj.payload.text || 'no text'));
              obj._sent = true;
              deferred.resolve('ACK');
              messageQueue = messageQueue.filter(function(a){ return !a._sent;});
            });
            setTimeout(function(){
              deferred.reject('timeout');
            }, 5000);
          });
        }
        return $q.all(deferreds);
      }

      var sendMsg = function(message, messageId, chatId, digitalAsset, location) {
        var deferred = $q.defer(),
          payload = {};
        if (message) payload.text = message;
        if (digitalAsset) {
          //payload.digitalAssetId = digitalAsset.id;
          //payload.digitalAssetType = digitalAsset.type;
          payload.digitalAsset = digitalAsset;
        }
        if (location) {
          payload.location = location;
        }

        send(messageId, chatId, 'message', 'sent', payload, 'profileId');
        deferred.resolve('sent');
        return deferred.promise;
      };

      var deleteMsg = function(messageId, chatId, digitalAssets) {
        //{ chatId: chat.currentChatID, scope: 'message', type: 'deleted', messageId: id}
        send(messageId, chatId, 'message', 'delete', {digitalAsset: (digitalAssets && digitalAssets[0] && _.pick(['id','type'], digitalAssets[0]))});
      };

      var wipeChat = function(chatId, from) {
        send(null, chatId, 'chat', 'wipe', null, from);
      };

      var messageAsRead = function(messageId, chatId) {
        // { chatId: chat.currentChatID, scope: 'message', from: {profileId: app.profileID, firstName: app.userData.firstName}, type: 'read', messageId: id}
        //console.log('websocketData.messageAsRead: ' + messageId);
        send(messageId, chatId, 'message', 'read');
      };

      var joinChat = function(chatId) {
        // { chatId: chatId, scope: 'chat', from: app.profileID, type: 'joined'}
        send(null, chatId, 'chat', 'joined', null, 'profileId');
      };

      var typing = function(chatId, from){
        // {chatId: "543499ce20ccb2202428a943", scope: "chat", type: "typing"}
        send(null, chatId, 'chat', 'typing', null, from);
      };

      function fire(event, data){
        $rootScope.$broadcast(event, data);
      };

      return {
        init: init,
        destruct: destruct,
        joinChat: joinChat,
        sendMsg: sendMsg,
        deleteMsg: deleteMsg,
        wipeChat: wipeChat,
        messageAsRead: messageAsRead,
        typing: typing,
        ready: function() {
          if (!ready) {
            ready = $q.defer();
          }
          return ready.promise;
        },
        _websocket: function(){
          return websocket;
        },
        _messageQueue: function(){
          return messageQueue;
        }
      };
    };

    return websocketData;
  });
}(define, io));